@@ -1,3 +1,4 @@
+/* SPDX-License-Identifier: MIT OR GPL-3.0-only */
 /* resolver.c
  * strophe XMPP client library -- DNS resolver
  *
@@ -6,7 +7,7 @@
  *  This software is provided AS-IS with no warranty, either express
  *  or implied.
  *
- *  This program is dual licensed under the MIT and GPLv3 licenses.
+ *  This program is dual licensed under the MIT or GPLv3 licenses.
  */
 
 /** @file
@@ -91,6 +92,25 @@ void resolver_shutdown(void)
 #endif
 }
 
+resolver_srv_rr_t *resolver_srv_rr_new(xmpp_ctx_t *ctx,
+                                       const char *host,
+                                       unsigned short port,
+                                       unsigned short prio,
+                                       unsigned short weight)
+{
+    resolver_srv_rr_t *rr = strophe_alloc(ctx, sizeof(*rr));
+    if (rr) {
+        memset(rr, 0, sizeof(*rr));
+        rr->port = port;
+        rr->priority = prio;
+        rr->weight = weight;
+        if (host) {
+            snprintf(rr->target, sizeof(rr->target), "%s", host);
+        }
+    }
+    return rr;
+}
+
 static void resolver_srv_list_sort(resolver_srv_rr_t **srv_rr_list)
 {
     resolver_srv_rr_t *rr_head;
@@ -158,6 +178,10 @@ int resolver_srv_lookup_buf(xmpp_ctx_t *ctx,
     set = resolver_ares_srv_lookup_buf(ctx, buf, len, srv_rr_list);
 #else
     set = resolver_raw_srv_lookup_buf(ctx, buf, len, srv_rr_list);
+    if (set != XMPP_DOMAIN_FOUND && *srv_rr_list != NULL) {
+        resolver_srv_free(ctx, *srv_rr_list);
+        *srv_rr_list = NULL;
+    }
 #endif
     resolver_srv_list_sort(srv_rr_list);
 
@@ -434,17 +458,19 @@ static int resolver_raw_srv_lookup_buf(xmpp_ctx_t *ctx,
         rdlength = xmpp_ntohs_ptr(&buf[j + 8]);
         j += 10;
         if (type == MESSAGE_T_SRV && class == MESSAGE_C_IN) {
-            rr = strophe_alloc(ctx, sizeof(*rr));
-            rr->next = *srv_rr_list;
-            rr->priority = xmpp_ntohs_ptr(&buf[j]);
-            rr->weight = xmpp_ntohs_ptr(&buf[j + 2]);
-            rr->port = xmpp_ntohs_ptr(&buf[j + 4]);
-            name_len = message_name_get(buf, len, j + 6, rr->target,
-                                        sizeof(rr->target));
-            if (name_len > 0)
-                *srv_rr_list = rr;
-            else
-                strophe_free(ctx, rr); /* skip broken record */
+            rr = resolver_srv_rr_new(ctx, NULL, 0, 0, 0);
+            if (rr) {
+                rr->next = *srv_rr_list;
+                rr->priority = xmpp_ntohs_ptr(&buf[j]);
+                rr->weight = xmpp_ntohs_ptr(&buf[j + 2]);
+                rr->port = xmpp_ntohs_ptr(&buf[j + 4]);
+                name_len = message_name_get(buf, len, j + 6, rr->target,
+                                            sizeof(rr->target));
+                if (name_len > 0)
+                    *srv_rr_list = rr;
+                else
+                    strophe_free(ctx, rr); /* skip broken record */
+            }
         }
         j += rdlength;
     }
